/* Copyright (C) 2023 Facility for Antiproton and Ion Research in Europe, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Dominik Smith [committer] */

#include "Hitfind.h"

#include <chrono>

#include "log.hpp"
#include "util/TimingsFormat.h"

using namespace std;
using fles::Subsystem;

namespace cbm::algo::tof
{
  // -----   Constructor   ------------------------------------------------------
  Hitfind::Hitfind(tof::HitfindSetup setup) : fNbSm(setup.NbSm), fNbRpc(setup.NbRpc), fStorDigi(fNbSm.size())
  {

    // Create one algorithm per RPC for TOF and configure it with parametersa
    for (uint32_t SmType = 0; SmType < fNbSm.size(); SmType++) {

      int32_t NbSm  = fNbSm[SmType];
      int32_t NbRpc = fNbRpc[SmType];
      fStorDigi[SmType].resize(NbSm * NbRpc);

      for (int32_t Sm = 0; Sm < NbSm; Sm++) {
        for (int32_t Rpc = 0; Rpc < NbRpc; Rpc++) {

          auto par = std::make_unique<cbm::algo::tof::ClusterizerRpcPar>();

          HitfindSetup::Rpc rpcPar = setup.rpcs[SmType][Sm * NbRpc + Rpc];
          par->fDeadStrips         = rpcPar.deadStrips;
          par->fPosYMaxScal        = rpcPar.posYMaxScal;
          par->fdMaxTimeDist       = rpcPar.maxTimeDist;
          par->fdMaxSpaceDist      = rpcPar.maxSpaceDist;
          par->fSigVel             = rpcPar.sigVel;
          par->fCPTOffYBinWidth    = rpcPar.CPTOffYBinWidth;
          par->fCPTOffY            = rpcPar.CPTOffY;
          par->fCPTOffYRange       = rpcPar.CPTOffYRange;
          par->fTimeRes            = rpcPar.timeRes;

          //geometry information, can in principle be done channel-wise but not needed for now
          const double* tra_ptr = rpcPar.cell.translation.data();
          const double* rot_ptr = rpcPar.cell.rotation.data();

          //channel parameters
          int32_t NbChan = rpcPar.chanPar.size();
          par->fChanPar.resize(NbChan);

          for (int32_t Ch = 0; Ch < NbChan; Ch++) {
            HitfindSetup::Channel chanPar   = rpcPar.chanPar[Ch];
            par->fChanPar[Ch].address       = chanPar.address;
            par->fChanPar[Ch].cell.pos      = ROOT::Math::XYZVector(tra_ptr[0], tra_ptr[1], tra_ptr[2]);
            par->fChanPar[Ch].cell.rotation = ROOT::Math::Rotation3D(&rot_ptr[0], &rot_ptr[9]);
            par->fChanPar[Ch].cell.sizeX    = rpcPar.cell.sizeX;
            par->fChanPar[Ch].cell.sizeY    = rpcPar.cell.sizeY;
          }
          fAlgo[SmType].emplace(std::make_pair(Sm * NbRpc + Rpc, tof::Clusterizer(std::move(*par))));
        }
      }
    }
    L_(info) << "--- Configured hitfinder algorithms for TOF.";
    L_(info) << "==================================================";
  }
  // ----------------------------------------------------------------------------


  // -----   Execution   -------------------------------------------------------
  Hitfind::resultType Hitfind::operator()(gsl::span<CbmTofDigi> digiIn)
  {
    xpu::push_timer("TofHitfind");
    xpu::t_add_bytes(digiIn.size_bytes());

    // --- Output data
    resultType result = {};

    auto& clusterTs = std::get<0>(result);
    auto& monitor   = std::get<1>(result);
    auto& digiInd   = std::get<2>(result);  // digi indices

    // Loop over the digis array and store the Digis in separate vectors for
    // each RPC modules
    for (size_t idigi = 0; idigi < digiIn.size(); idigi++) {
      CbmTofDigi* pDigi = &(digiIn[idigi]);

      // These are doubles in the digi class
      const double SmType = pDigi->GetType();
      const double Sm     = pDigi->GetSm();
      const double Rpc    = pDigi->GetRpc();
      const int NbRpc     = fNbRpc[SmType];
      if (SmType >= fNbSm.size() || Sm * NbRpc + Rpc >= fNbSm[SmType] * NbRpc) {
        // Error already counted for monitoring during Digis calibration, so just discard here
        continue;
      }
      fStorDigi[SmType][Sm * NbRpc + Rpc].emplace_back(*pDigi, idigi);
    }

    std::vector<Hit> clustersFlat;
    std::vector<size_t> chanSizes;   // nClusters per channel
    std::vector<u32> chanAddresses;  // channel addresses

    // ---  RPC loop
    for (uint32_t SmType = 0; SmType < fNbSm.size(); SmType++) {
      const uint32_t NbRpc = fNbRpc[SmType];
      const uint32_t NbSm  = fNbSm[SmType];
      for (uint32_t Sm = 0; Sm < NbSm; Sm++) {
        for (uint32_t Rpc = 0; Rpc < NbRpc; Rpc++) {

          // Get digis
          std::vector<std::pair<CbmTofDigi, int32_t>>& digiExp = fStorDigi[SmType][Sm * NbRpc + Rpc];

          // Build clusters
          auto rpcresult              = fAlgo[SmType][Sm * NbRpc + Rpc](digiExp);
          std::vector<Hit>& clusters  = std::get<0>(rpcresult);  // Hits
          std::vector<size_t>& sizes  = std::get<1>(rpcresult);  // nClusters per channel
          std::vector<u32>& addresses = std::get<2>(rpcresult);  // channel addresses
          std::vector<i32>& indices   = std::get<3>(rpcresult);  // digi indices

          // Append clusters to output
          clustersFlat.insert(clustersFlat.end(), std::make_move_iterator(clusters.begin()),
                              std::make_move_iterator(clusters.end()));

          // Store hw address of partition
          chanAddresses.insert(chanAddresses.end(), std::make_move_iterator(addresses.begin()),
                               std::make_move_iterator(addresses.end()));

          // store partition size
          chanSizes.insert(chanSizes.end(), std::make_move_iterator(sizes.begin()),
                           std::make_move_iterator(sizes.end()));

          // store digi indices
          digiInd.insert(digiInd.end(), std::make_move_iterator(indices.begin()),
                         std::make_move_iterator(indices.end()));

          // Clear digi storage
          digiExp.clear();
        }
      }
    }

    // Monitoring
    monitor.fTime     = xpu::pop_timer();
    monitor.fNumDigis = digiIn.size();
    monitor.fNumHits  = clustersFlat.size();

    // Create ouput vector
    clusterTs = PartitionedVector(std::move(clustersFlat), chanSizes, chanAddresses);

    return result;
  }
  // ----------------------------------------------------------------------------

}  // namespace cbm::algo::tof
