/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// @file   CbmCaInputQaBase.h
/// @date   13.01.2023
/// @brief  Class providing basic functionality for CA input QA-tasks
/// @author S.Zharko <s.zharko@gsi.de>

#ifndef CbmCaInputQaBase_h
#define CbmCaInputQaBase_h 1

#include "CbmCaHitQaData.h"
#include "CbmCaInputQaBase.h"
#include "CbmL1DetectorID.h"
#include "CbmMCDataManager.h"
#include "CbmQaTask.h"

#include "TMath.h"

#include <set>
#include <unordered_map>
#include <vector>

class CbmMatch;
class CbmMCEventList;
class CbmMCDataArray;
class CbmMCDataManager;
class CbmMCTrack;
class CbmStsHit;
class CbmStsPoint;
class CbmTrackingDetectorInterfaceBase;
class CbmTimeSlice;
class FairMCPoint;
class TClonesArray;
class TH1F;
class TH2F;
class TProfile;
class TProfile2D;

namespace
{
  using cbm::algo::ca::EDetectorID;
}

/// A QA-task class, which provides assurance of MuCh hits and geometry
template<EDetectorID DetID>
class CbmCaInputQaBase : public CbmQaTask {
protected:
  using Point_t = cbm::ca::PointTypes_t::at<DetID>;  ///< Point type for detector ID
  using Hit_t   = cbm::ca::HitTypes_t::at<DetID>;    ///< Hit type for detector ID

public:
  /// @brief Constructor from parameters
  /// @param  name      Name of the task
  /// @param  verbose   Verbose level
  /// @param  isMCUsed  Flag, whether MC information is available for this task
  CbmCaInputQaBase(const char* name, int verbose, bool isMCUsed);

  /// @brief Destructor
  virtual ~CbmCaInputQaBase() = default;


  /// @brief Sets maximum allowed deviation of pull distribution mean from 0
  /// @param devThrsh  Deviation threshold (- devThrsh, + devThrsh)
  void SetPullMeanDeviationThrsh(double devThrsh) { fPullMeanThrsh = devThrsh; }

  /// @brief Sets maximum allowed deviation of pull distribution width from unity
  /// @param devThrsh  Deviation threshold (1 - devThrsh, 1 + devThrsh)
  void SetPullWidthDeviationThrsh(double devThrsh) { fPullWidthThrsh = devThrsh; }

  /// @brief Sets maximum allowed deviation of residual mean from zero (-devThrsh * rms, +devThrsh * rms)
  /// @param devThrsh  Deviation threshold in sigmas of residual distribution
  void SetResidualMeanDeviationThrsh(double devThrsh) { fResMeanThrsh = devThrsh; }

  /// @brief Sets hit efficiency analysis parameters: efficiency fit range
  /// @param thrsh    Threshold for efficiency in a given range
  /// @param loRange  Lower bound of the integration range
  /// @param upRange  Upper limit of the integration range
  void SetEfficiencyThrsh(double thrsh, double loRange, double upRange)
  {
    LOG(info) << "Call: SetEfficiencyThrsh";
    fEffThrsh    = thrsh;
    fEffRange[0] = loRange;
    fEffRange[1] = upRange;
    LOG(info) << fEffRange[0] << ' ' << fEffRange[1];
  }

  /// @brief Sets maximum allowed difference between z-position of hit and station
  /// @param  dz  Difference between station and hit position z components [cm]
  void SetMaxDiffZStationHit(double dz) { fMaxDiffZStHit = dz; }

  /// @brief Sets lower momentum threshold
  /// @param mom Minimum absolute value of momentum
  void SetMinMomentum(double mom) { fMinMomentum = mom; }

protected:
  // ********************************************
  // ** Virtual method override from CbmQaTask **
  // ********************************************

  /// \brief  Method to check, if the QA results are acceptable
  /// \return Map of checks: key - name of the check, value - result of the check
  std::map<std::string, bool> Check();

  /// @brief Initializes data branches
  InitStatus InitDataBranches();

  /// @brief Initializes canvases
  InitStatus InitCanvases();

  /// @brief Initializes histograms
  InitStatus InitHistograms();

  /// @brief Fills histograms per event or time-slice
  void FillHistograms();

  /// @brief Fills histograms per hit
  virtual void FillHistogramsPerHit() {}

  /// @brief Fills histograms per MC point
  virtual void FillHistogramsPerPoint() {}

  /// @brief De-initializes histograms
  void DeInit();

  /// @brief Defines parameter for a derived class
  virtual void DefineParameters() = 0;

  /// @brief Checks ranges for mean and standard deviation
  /// @return  False, if variable exits the range
  bool CheckRangePull(TH1* h)
  {
    return CbmQaTask::CheckRange(h, fPullMeanThrsh, 1. - fPullWidthThrsh, 1. + fPullWidthThrsh);
  }

  /// @brief Returns a pointer to current hit QA data object
  ///
  /// The hit QA data object is filled on each step for hit reconstruction and available inside
  /// functions FillHit() and FillMCPoint()
  const cbm::ca::HitQaData& GetHitQaData() const { return fHitQaData; }

  /// @brief checks if the track at mc point passes the track selection cuts
  /// @return true when selected
  bool IsTrackSelected(const CbmMCTrack* track, const Point_t* point) const;

  /// @struct ResidualFitResult
  /// @brief  Stores fit residuals result
  struct ResidualFitResult {
    double mean = 0;  ///< mean of the distribution
    double lo   = 0;  ///< lower limit for the mean
    double hi   = 0;  ///< higher limit for the mean
  };

  // ******************
  // **  Parameters  **
  // ******************

  static constexpr double kNAN = std::numeric_limits<double>::signaling_NaN();

  //TODO: remove check of residuals
  // ----- QA constants (thresholds, ranges etc.)
  double fResMeanThrsh   = kNAN;  ///< Maximum allowed deviation of residual mean from zero [sigma]
  double fPullMeanThrsh  = kNAN;  ///< Maximum allowed deviation of pull mean from zero
  double fPullWidthThrsh = kNAN;  ///< Maximum allowed deviation of pull width from unity
  double fEffThrsh       = kNAN;  ///< Threshold for hit efficiency in the selected range
  double fMaxDiffZStHit  = kNAN;  ///< Maximum allowed difference between z-position of hit and station [cm]
  double fMinMomentum    = kNAN;  ///< Minimum momentum of particle [GeV/c]

  std::array<double, 2> fEffRange = {kNAN, kNAN};  ///< Range for hit efficiency approximation

  // ----- Track selection for efficiencies and pulls
  bool fTrackSelectionPrimary    = true;  ///< must the track come from the primary vertex
  double fTrackSelectionMomentum = kNAN;  ///< track momentum GeV when it exits the tracking station
  double fTrackSelectionAngle    = kNAN;  ///< track incl. angle [grad] when it exits the station

  // ----- Histogram binning parameters
  int fNbins  = 200;  ///< General number of bins
  int fNbinsZ = 800;  ///< Number of bins for z axis in overall views

  std::array<double, 2> fRHitDx = {kNAN, kNAN};  ///< Range for hit x coordinate error [cm]
  std::array<double, 2> fRHitDy = {kNAN, kNAN};  ///< Range for hit y coordinate error [cm]
  std::array<double, 2> fRHitDu = {kNAN, kNAN};  ///< Range for hit u coordinate error [cm]
  std::array<double, 2> fRHitDv = {kNAN, kNAN};  ///< Range for hit v coordinate error [cm]
  std::array<double, 2> fRHitDt = {kNAN, kNAN};  ///< Range for hit time error [ns]

  std::array<double, 2> fRResX = {kNAN, kNAN};  ///< Range for residual of x coordinate [cm]
  std::array<double, 2> fRResY = {kNAN, kNAN};  ///< Range for residual of y coordinate [cm]
  std::array<double, 2> fRResU = {kNAN, kNAN};  ///< Range for residual of u coordinate [cm]
  std::array<double, 2> fRResV = {kNAN, kNAN};  ///< Range for residual of v coordinate [cm]
  std::array<double, 2> fRResT = {kNAN, kNAN};  ///< Range for residual of time [ns]

  std::array<double, 2> fRangeDzHitPoint = {-0.05, 0.05};  ///< Range for z-pos difference of the hit and point [cm]

  // NOTE: Pull binning is fixed by convention, since it is used for hit finder calibrations. Please,
  //       do not modify!
  static constexpr int kNbinsPull   = 200;
  static constexpr double kRPull[2] = {-10., 10.};  ///< Range for pull histograms coordinate

  std::vector<double> frXmin;  ///< Range for hit x coordinate [cm]
  std::vector<double> frXmax;  ///< Range for hit x coordinate [cm]

  std::vector<double> frYmin;  ///< Range for hit y coordinate [cm]
  std::vector<double> frYmax;  ///< Range for hit y coordinate [cm]

  std::vector<double> frZmin;  ///< Range for hit z coordinate [cm]
  std::vector<double> frZmax;  ///< Range for hit z coordinate [cm]

  // *********************
  // **  Data branches  **
  // *********************

  CbmTrackingDetectorInterfaceBase* fpDetInterface = nullptr;  ///< Instance of detector interface

  CbmTimeSlice* fpTimeSlice = nullptr;  ///< Pointer to current time-slice

  TClonesArray* fpHits     = nullptr;  ///< Array of hits
  TClonesArray* fpClusters = nullptr;  ///< Array of hit clusters

  CbmMCDataManager* fpMCDataManager = nullptr;  ///< MC data manager
  CbmMCEventList* fpMCEventList     = nullptr;  ///< MC event list
  CbmMCDataArray* fpMCTracks        = nullptr;  ///< Array of MC tracks
  CbmMCDataArray* fpMCPoints        = nullptr;  ///< Array of MC points

  TClonesArray* fpHitMatches = nullptr;  ///< Array of hit matches


  // ----- Histograms


  // NOTE: the last element of each vector stands for integral distribution over all stations

  // ----- Additional histograms
  // Hit occupancy
  std::vector<TH2F*> fvph_hit_xy;  ///< hit y vs x in different stations
  std::vector<TH2F*> fvph_hit_zx;  ///< hit x vs z in different stations
  std::vector<TH2F*> fvph_hit_zy;  ///< hit y vs z in different stations

  std::vector<TH1F*> fvph_hit_station_delta_z;  ///< Difference between station and hit z positions [cm]

  // Hit errors
  std::vector<TH1F*> fvph_hit_dx;
  std::vector<TH1F*> fvph_hit_dy;
  std::vector<TH1F*> fvph_hit_du;
  std::vector<TH1F*> fvph_hit_dv;
  std::vector<TH1F*> fvph_hit_kuv;
  std::vector<TH1F*> fvph_hit_dt;

  // MC points occupancy
  std::vector<TH1F*> fvph_n_points_per_hit;  ///< number of points per one hit in different stations

  std::vector<TH2F*> fvph_point_xy;  ///< point y vs x in different stations
  std::vector<TH2F*> fvph_point_zx;  ///< point x vs z in different stations
  std::vector<TH2F*> fvph_point_zy;  ///< point y vs z in different stations

  std::vector<TH1F*> fvph_point_hit_delta_z;  ///< difference between z of hit and MC point in different stations

  // Residuals
  std::vector<TH1F*> fvph_res_x;  ///< residual for x coordinate in different stations
  std::vector<TH1F*> fvph_res_y;  ///< residual for y coordinate in different stations
  std::vector<TH1F*> fvph_res_u;  ///< residual for u coordinate in different stations
  std::vector<TH1F*> fvph_res_v;  ///< residual for v coordinate in different stations
  std::vector<TH1F*> fvph_res_t;  ///< residual for t coordinate in different stations

  std::vector<TH2F*> fvph_res_x_vs_x;  ///< residual for x coordinate in different stations
  std::vector<TH2F*> fvph_res_y_vs_y;  ///< residual for y coordinate in different stations
  std::vector<TH2F*> fvph_res_u_vs_u;  ///< residual for u coordinate in different stations
  std::vector<TH2F*> fvph_res_v_vs_v;  ///< residual for v coordinate in different stations
  std::vector<TH2F*> fvph_res_t_vs_t;  ///< residual for t coordinate in different stations

  // Pulls
  std::vector<TH1F*> fvph_pull_x;  ///< pull for x coordinate in different stations
  std::vector<TH1F*> fvph_pull_y;  ///< pull for y coordinate in different stations
  std::vector<TH1F*> fvph_pull_u;  ///< pull for u coordinate in different stations
  std::vector<TH1F*> fvph_pull_v;  ///< pull for v coordinate in different stations
  std::vector<TH1F*> fvph_pull_t;  ///< pull for t coordinate in different stations

  std::vector<TH2F*> fvph_pull_x_vs_x;  ///< pull for x coordinate in different stations
  std::vector<TH2F*> fvph_pull_y_vs_y;  ///< pull for y coordinate in different stations
  std::vector<TH2F*> fvph_pull_u_vs_u;  ///< pull for u coordinate in different stations
  std::vector<TH2F*> fvph_pull_v_vs_v;  ///< pull for v coordinate in different stations
  std::vector<TH2F*> fvph_pull_t_vs_t;  ///< pull for t coordinate in different stations

  // Hit efficiencies
  std::vector<TProfile2D*> fvpe_reco_eff_vs_xy;  ///< Efficiency of hit reconstruction vs. x and y coordinates of MC
  std::vector<TH1F*> fvph_reco_eff;  ///< Distribution of hit reconstruction efficiency in bins of fvpe_reco_eff_vs_xy

  // FIXME: change to private
protected:
  cbm::ca::HitQaData fHitQaData {};  ///< Current hit QA data object
};

#endif  // CbmCaInputQaBase_h
